@@ -1,32 +1,33 @@
-#!/usr/bin/env python
-"""
-Copyright (c) 2013-2015 Fraunhofer FKIE
-
-Redistribution and use in source and binary forms, with or without
-modification, are permitted provided that the following conditions
-are met:
-
- * Redistributions of source code must retain the above copyright
-   notice, this list of conditions and the following disclaimer.
- * Redistributions in binary form must reproduce the above copyright
-   notice, this list of conditions and the following disclaimer in the
-   documentation and/or other materials provided with the distribution.
- * Neither the name of the Fraunhofer organization nor the names of its
-   contributors may be used to endorse or promote products derived from
-   this software without specific prior written permission.
-
-THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
-IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
-TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
-PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
-HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
-TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
-PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
-LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
-NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
-SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-"""
+# coding=utf-8
+#
+# catkin_lint
+# Copyright (c) 2013-2018 Fraunhofer FKIE
+#
+# Redistribution and use in source and binary forms, with or without
+# modification, are permitted provided that the following conditions
+# are met:
+#
+#  * Redistributions of source code must retain the above copyright
+#    notice, this list of conditions and the following disclaimer.
+#  * Redistributions in binary form must reproduce the above copyright
+#    notice, this list of conditions and the following disclaimer in the
+#    documentation and/or other materials provided with the distribution.
+#  * Neither the name of the Fraunhofer organization nor the names of its
+#    contributors may be used to endorse or promote products derived from
+#    this software without specific prior written permission.
+#
+# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
+# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
+# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
+# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
+# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
+# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
+# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
+# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+import re
 
 message_list = {
     "ARGUMENT_ERROR":
@@ -52,10 +53,15 @@ message_list = {
     "DUPLICATE_FIND":
     ("duplicate find_package(%(pkg)s)",
         """\
-        The find_package() searches for a dependency and caches the result.
-        A second call will be silently ignored. In particular, if you
-        specify different arguments to the second call, those will not have
-        any effect at all, which is most likely not what you want.
+        You called find_package() more than once for a particular package,
+        which is not needed except for very specific, advanced circumstances.
+        """),
+    "SHADOWED_FIND":
+    ("call to find_package(%(pkg)s) shadows previously selected components",
+        """\
+        You have more than one find_package() call for a package, and the
+        COMPONENTS list of the later call does not include a previously
+        chosen component.
         """),
     "MISSING_FIND":
     ("missing find_package(%(pkg)s)",
@@ -69,6 +75,13 @@ message_list = {
         backwards compatibility reasons, both names should also be identical
         to the name of the source folder that contains the package.
         """),
+    "PACKAGE_PATH_NAME":
+    ("package path name '%(path)s' differs from package name",
+        """\
+        Your package resides in a folder that has a different name than the
+        package itself. This is confusing and might break the assumptions of
+        some tools.
+        """),
     "LITERAL_PROJECT_NAME":
     ("use ${PROJECT_NAME} instead of '%(name)s'",
         """\
@@ -88,12 +101,6 @@ message_list = {
         Catkin macros cannot be called before catkin has been configured with
         find_package(catkin).
         """),
-    "DEPRECATED_ROSBUILD":
-    ("%(cmd)s() is deprecated",
-        """\
-        This construct was intended to facilitate the migration from
-        Rosbuild to Catkin. It is deprecated and should not be used any more.
-        """),
     "DEPRECATED_CMD":
     ("%(old_cmd)s() is deprecated, use %(new_cmd)s() instead",
         """\
@@ -139,14 +146,9 @@ message_list = {
     "UNKNOWN_PACKAGE":
     ("unknown package '%(pkg)s'",
         """\
-        You have listed a package which is neither a catkin package nor a known system
-        dependency.
-        """),
-    "MISSING_CATKIN_DEPEND":
-    ("%(type)s_depend '%(pkg)s' is not listed in catkin_package()",
-        """\
-        You have specified a catkin run dependency but failed to list
-        it in the CATKIN_DEPENDS stanza of the catkin_package() call.
+        You are referring to a package which seems to be neither a catkin package
+        nor a known system dependency. You may have misspelled the name, or your
+        rosdep database needs to be refreshed with "rosdep update".
         """),
     "INVALID_META_COMMAND":
     ("%(cmd)s() is not allowed in meta packages",
@@ -161,14 +163,14 @@ message_list = {
         As meta packages do neither build nor test anything, the only valid
         dependency type is the run_depend.
         """),
-    "CATKIN_PKG_VS_META":
+    "WRONG_CATKIN_PACKAGE":
     ("catkin_package() in meta package",
         """\
         Meta packages use the catkin_metapackage() command to declare a
         meta package. This performs additional checks and ensures that all
         requirements are met.
         """),
-    "CATKIN_META_VS_PKG":
+    "WRONG_CATKIN_METAPACKAGE":
     ("catkin_metapackage() in regular package",
         """\
         The catkin_metapackage() command signals your intent to declare
@@ -209,26 +211,58 @@ message_list = {
         i.e. the first line has to start with '#!' and needs to name
         the full path to the %(interpreter)s executable.
         """),
+    "EXTERNAL_FILE":
+    ("%(cmd)s() uses file '%(file)s' which is not in package",
+        """\
+        This catkin command uses a file which lies outside of the package
+        source folder. While this may work in your particular setup, you
+        cannot assume file locations in general. Use find_file() to detect
+        external locations insteed.
+        """),
     "MISSING_FILE":
     ("%(cmd)s() needs missing file '%(file)s'",
         """\
         This catkin command processes a particular file which is missing
         from the package source folder.
         """),
+    "EXTERNAL_DIRECTORY":
+    ("%(cmd)s() uses directory '%(directory)s' which is not in package",
+        """\
+        This catkin command uses a directory which lies outside of the package
+        source folder. While this may work in your particular setup, you
+        cannot assume file locations in general. Use find_path() to detect
+        external locations insteed.
+        """),
     "MISSING_DIRECTORY":
     ("%(cmd)s() needs missing directory '%(directory)s'",
         """\
         This catkin command processes a particular directory which is missing
         from the package source folder.
         """),
-    "INSTALL_DESTINATION":
-    ("install(%(type)s ... %(dest)s) is not one of the ${CATKIN_*_DESTINATION}s",
+    "BAD_INTERFACE_DIRECTORY":
+    ("%(cmd)s() has foreign %(scope)s path '%(directory)s'",
         """\
-        Catkin provides a number of standard variables to specify
-        installation folders. You should use those to ensure that your
+        All properties of targets are exported hardcoded, so you must not export
+        any public paths if they do not belong to your package. Otherwise, they
+        will potentially break if your CMake targets are imported again on a
+        different machine. Use the PRIVATE scope instead.
+        """),
+    "WRONG_INSTALL_DESTINATION":
+    ("install(%(type)s ... %(dest)s) does not install to ${CATKIN_INSTALL_PREFIX}",
+        """\
+        Your package installs one or more files to an unexpected location.
+        Catkin provides a number of standard variables ${CATKIN_*_DESTINATION}
+        to specify installation folders. You should use those to ensure that your
         package will continue to work if the file system layout is
         changed in the future.
         """),
+    "WRONG_BIN_INSTALL_DESTINATION":
+    ("executable file is not installed to bin destination",
+        """\
+        Your package installs one or more files to an unexpected location.
+        Executable files should end up in either ${CATKIN_GLOBAL_BIN_DESTINATION}
+        or ${CATKIN_PACKAGE_BIN_DESTINATION}.
+        """),
     "UNUSED_DEPEND":
     ("unused %(type)s_depend on '%(pkg)s'",
         """\
@@ -241,6 +275,13 @@ message_list = {
         Your package uses features of another package but you
         failed to list this dependency in your package.xml
         """),
+    "LAUNCH_DEPEND":
+    ("launch configuration needs %(type)s_depend on '%(pkg)s'",
+        """\
+        Your package refers to another package in one of its
+        launch files, but you do not have this dependency in your
+        package.xml
+        """),
     "UNCONFIGURED_BUILD_DEPEND":
     ("unconfigured build_depend on '%(pkg)s'",
         """\
@@ -248,16 +289,22 @@ message_list = {
         call find_package() nor have it listed as catkin component in
         the find_package(catkin) call.
         """),
+    "MISCONFIGURED_CATKIN_PACKAGE":
+    ("misconfigured catkin package '%(pkg)s'",
+        """\
+        You use an unsupported way to include a catkin package in your build. Even though this
+        might work in your particular case, you should use the proper
+        find_package() mechanism to make sure that all relevant CMake macros will be run.
+        """),
     "WRONG_DEPEND":
     ("%(wrong_type)s_depend '%(pkg)s' should be a %(right_type)s_depend",
         """\
-        You have listed a package as the wrong dependency type. build_depends are needed
-        to build your package (as in compile the declared executables and libraries).
-        run_depends are needed at runtime to run the nodes or use the libraries
-        and exported headers in other projects. buildtool_depends are significant only
-        for cross-compiling; in that case, buildtool_depends are host architecture (and run
-        during the build process) while build_depends are target architecture (and are
-        linked against). test_depends are additional run_depends which only apply to unit tests.
+        You have listed a package as the wrong dependency type. The catkin build system
+        defines a number of different dependencies for building a package, running its nodes
+        or linking against its libraries. As a general rule, buildtool_depends are needed
+        to build the package, build_depends are non-transitive build dependencies which are not visible
+        to users of your package, build_export_depends are publicly visible transitive dependencies,
+        exec_depends are needed at runtime, and test_depends are for unit tests only.
         """),
     "UNUSED_GENERATE_MSG":
     ("generate_messages() called but no message declared",
@@ -274,55 +321,50 @@ message_list = {
         and add_action_files() respectively.
         """),
     "SYSTEM_AS_CATKIN_DEPEND":
-    ("catkin_package() lists '%(pkg)s' as catkin package but it is not",
+    ("catkin_package() dependency '%(pkg)s' belongs in DEPENDS",
         """\
         In your catkin_package() call, you have listed a system dependency in the
         CATKIN_DEPENDS stanza, but it belongs in the DEPENDS stanza instead.
         """),
     "CATKIN_AS_SYSTEM_DEPEND":
-    ("catkin_package() lists '%(pkg)s' as system package but it is not",
+    ("catkin_package() dependency '%(pkg)s' belongs in CATKIN_DEPENDS",
         """\
         In your catkin_package() call, you have listed a catkin package in the
         DEPENDS stanza, but it belongs in the CATKIN_DEPENDS stanza instead.
         """),
     "UNCONFIGURED_SYSTEM_DEPEND":
-    ("catkin_package() lists unconfigured system package '%(pkg)s'",
+    ("catkin_package() dependency '%(pkg)s' is not configured properly",
         """\
         In order to export a system package as dependency, you must either
         call find_package(%(pkg)s) first or initialize the %(pkg)s_INCLUDE_DIRS and
         %(pkg)s_LIBRARIES variables manually.
         """),
-    "MISSING_BUILD_INCLUDE":
+    "UNUSED_INCLUDE_PATH":
     ("include path '%(path)s' is exported but not used for the build",
         """\
         You have listed an include path in the INCLUDE_DIRS stanza of the
         catkin_package() command, but that path is not mentioned in any
         include_directories() call.
         """),
-    "DUPLICATE_BUILD_INCLUDE":
+    "DUPLICATE_INCLUDE_PATH":
     ("duplicate include path ${%(pkg)s_INCLUDE_DIRS}",
         """\
         Include paths of packages listed in the find_package(catkin) command are added implicitly
         by the ${catkin_INCLUDE_DIRS} variable. There is no need to add it a second time.
         """),
-    "AMBIGUOUS_BUILD_INCLUDE":
+    "AMBIGUOUS_INCLUDE_PATH":
     ("include paths '%(path)s' and '%(parent_path)s' are ambiguous",
         """\
         You have used two include paths where one is a parent of
         the other. Thus the same headers can be included with two different include paths
         which may confuse users. It is recommended that you keep your include paths consistent.
         """),
-    "MISSING_EXPORT_INCLUDE_PATH":
+    "MISSING_INCLUDE_PATH":
     ("exported include path '%(path)s' does not exist",
         """\
         You have listed an invalid include path in the INCLUDE_DIRS stanza of the
         catkin_package() command.
         """),
-    "MISSING_BUILD_INCLUDE_PATH":
-    ("build include path '%(path)s' does not exist",
-        """\
-        You have listed an invalid include path in the include_directories() command.
-        """),
     "EXTERNAL_INCLUDE_PATH":
     ("catkin_package() exports non-package include path",
         """\
@@ -332,19 +374,19 @@ message_list = {
         find_path(), and/or find_library() and add the dependency to the
         DEPENDS stanza.
         """),
-    "MISSING_CATKIN_INCLUDE":
+    "UNUSED_CATKIN_INCLUDE_DIRS":
     ("missing include_directories(${catkin_INCLUDE_DIRS})",
         """\
         You must add the catkin include paths to your include search list, or
         you might experience build failures.
         """),
-    "MISSING_INSTALL_INCLUDE":
+    "UNINSTALLED_INCLUDE_PATH":
     ("catkin_package() exports package include path that is not installed",
         """\
         Your package can be used from the devel space but cannot be installed
         properly, because the header files will not be copied to the proper location.
         """),
-    "MISSING_INSTALL_TARGET":
+    "UNINSTALLED_TARGET":
     ("target '%(target)s' is not installed",
         """\
         Your package can be used from the devel space but cannot be installed
@@ -370,18 +412,6 @@ message_list = {
         Your messages depend on another package which is neither find_package()'d
         nor listed as a component in the find_package(catkin) call.
         """),
-    "MISSING_MSG_DEPEND":
-    ("message dependency '%(pkg)s' is not listed as %(type)s_depend",
-        """\
-        Your messages depend on another package which is not listed as %(type)s_depend in
-        your package.xml
-        """),
-    "MISSING_MSG_CATKIN":
-    ("message dependency '%(pkg)s' is not listed in catkin_package()",
-        """\
-        Your messages depend on another package which is not in the CATKIN_DEPENDS
-        stanza of your catkin_package() call.
-        """),
     "UNINSTALLED_EXPORT_LIB":
     ("exported library '%(target)s' is not installed",
         """\
@@ -402,25 +432,18 @@ message_list = {
         library export function will break if the logical target name is not
         equal to the actual library name.
         """),
-    "SUGGEST_CATKIN_DEPEND":
-    ("package '%(pkg)s' should be listed in catkin_package()",
+    "MISSING_CATKIN_DEPEND":
+    ("package '%(pkg)s' must be in CATKIN_DEPENDS in catkin_package()",
         """\
-        Your package configures another package as build dependency, it is listed as
-        run_depend in your package.xml, and its name suggests
-        that it contains ROS messages. In that case, you must add it to the
-        CATKIN_DEPENDS stanza of your catkin_package()
+        You have a runtime dependency that
+        you must add to the CATKIN_DEPENDS stanza of your catkin_package().
         """),
     "UNDEFINED_TARGET":
-    ("exported target '%(target)s' is not defined",
+    ("referenced target '%(target)s' is not defined",
         """\
-        Your package provides a CMake target to other packages, but the listed
-        target is not defined at all.
-        """),
-    "UNDEFINED_INSTALL_TARGET":
-    ("installed target '%(target)s' is not defined",
-        """\
-        Your package installs a CMake target which is neither a library nor an
-        executable.
+        Your package installs or exports a CMake target which is not defined at all.
+        This could be a typo, or the target is implicitly defined by a macro that
+        is unknown to catkin_lint.
         """),
     "INVALID_TARGET_OUTPUT":
     ("target '%(target)s' has invalid characters in its output file name",
@@ -429,16 +452,6 @@ message_list = {
         characters in its name. You probably forget to call set_target_properties(... PROPERTIES
         OUTPUT_NAME ...)
         """),
-    "TARGET_NAME_COLLISION":
-    ("target name '%(target)s' might not be sufficiently unique",
-        """\
-        The CMake build system requires all target identifiers to be globally unique.
-        For this reason, it is highly recommended that you add the package name as in
-        '${PROJECT_NAME}_target' or '${PROJECT_NAME}/target'.
-        You can use set_target_properties(... PROPERTIES OUTPUT_NAME ...)
-        to give your target a different output file name (which does not have to
-        be unique if it is installed in a package-specific location).
-        """),
     "UNINSTALLED_DEPEND":
     ("target '%(export_target)s' depends on target '%(target)s' which is not installed",
         """\
@@ -481,13 +494,13 @@ message_list = {
         The ${prefix} variable is carefully overloaded to work with both
         devel space and install space and must be used in all <export plugin='...'> tags.
         """),
-    "PLUGIN_MISSING_FILE":
+    "MISSING_PLUGIN":
     ("%(export)s plugin refers to missing file '%(file)s'",
         """\
         A plugin declaration file which is listed in your package.xml is missing from
         the package source folder.
         """),
-    "PLUGIN_MISSING_INSTALL":
+    "UNINSTALLED_PLUGIN":
     ("%(export)s plugin file '%(file)s' is not installed to ${CATKIN_PACKAGE_SHARE_DESTINATION}",
         """\
         Your package can be used from the devel space but cannot be installed
@@ -520,12 +533,6 @@ message_list = {
         The FindXXX.cmake modules are intended to be included by the find_package()
         command.
         """),
-    "REDUNDANT_TEST_DEPEND":
-    ("redundant test_depend '%(pkg)s'",
-        """\
-        Test dependencies are additional dependencies for testing, so there is no
-        need to list any build or run dependency a second time.
-        """),
     "REDUNDANT_LIB_PREFIX":
     ("library output name '%(output)s' has redundant 'lib' prefix",
         """\
@@ -534,18 +541,6 @@ message_list = {
         set_target_properties(... PROPERTIES OUTPUT_NAME ...) to give your library a
         different file name without changing the target name.
         """),
-    "UNKNOWN_DEPEND":
-    ("unknown %(type)s_depend '%(pkg)s'",
-        """\
-        The specified dependency is neither a catkin package nor a known system dependency
-        from the rosdep database.
-        """),
-    "UNSUPPORTED_CMD":
-    ("unsupported command '%(cmd)s'",
-        """\
-        Your package uses CMake constructs which cannot be linted properly at this time.
-        Certain errors may go unnoticed while other errors may be false positives.
-        """),
     "EXTERNAL_SUBDIR":
     ("subdirectory %(subdir)s is not in package",
         """\
@@ -598,6 +593,11 @@ message_list = {
         An operating system error has occured. This is not a linting problem per se but
         might be caused by a missing or unreadable file.
         """),
+    "PARSE_ERROR":
+    ("parse error: %(msg)s",
+        """\
+        Your package has a malformed file that could not be processed for linting.
+        """),
     "ENDBLOCK_ARGS":
     ("extra arguments in %(cmd)s()",
         """\
@@ -653,4 +653,14 @@ message_list = {
 
 def msg(msg_id, **kwargs):
     text, explanation = message_list[msg_id]
+    explanation = re.sub(r"\s+", " ", explanation).strip()
     return (msg_id, text % kwargs, explanation % kwargs)
+
+
+def add_user_defined_msg(msg_id, text, explanation):
+    if msg_id in message_list:
+        old_text, old_explanation = message_list[msg_id]
+        if text != old_text or explanation != old_explanation:
+            raise ValueError("Conflicting definitions for message ID: %s" % msg_id)
+    else:
+        message_list[msg_id] = (text, explanation)
